#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#include <sys/types.h>
#include <sys/socket.h>

#include <inttypes.h>

#include <gsloop_types.h>
#include <gsloop_server.h>
#include <gsloop_con.h>

/**
 * make_socket_inet - create a new socket for PF_INET
 */
static int gsloop_inet_server_open(int		type,
				   uint16_t	port)
{
	struct sockaddr_in	name;
	int			new_sock;
	int			yes = 1;
	int			err;

	/* Create the socket.  */
	new_sock = socket(PF_INET, type, 0);
	if (new_sock == -1) {
		perror("socket");
		err = -errno;
		goto exit_0;
	}
	
	err = setsockopt(new_sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof (yes));
	if (err < 0) {
		perror ("setsockopt");
		err = -errno;
		goto exit_1;
	}

	/* Give the socket a name */
	memset(&name, 0, sizeof(name));
	name.sin_family	     = type;
	name.sin_port	     = htons(port);
	name.sin_addr.s_addr = htonl(INADDR_ANY);
	err = bind(new_sock, (struct sockaddr *)&name, sizeof(name));
	if (err < 0) {
		perror("bind");
		err = -errno;
		goto exit_2;
	}

	/* we are server, so listen to incoming connections */
	err = listen(new_sock, 8);	/* FIXME: backlog */
	if (err < 0) {
		perror("listen");
		err = -errno;
		goto exit_3;
	}

	return new_sock;

 exit_3:
 exit_2:
 exit_1:
	close(new_sock);
 exit_0:
	return err;
}




static int gsloop_inet_server_close(struct gsloop	*gsloop,
				    union gsloop_con	*gsloop_con)
{
	struct gsloop_server	*gsloop_server = &gsloop_con->server;
	int			err;

	if (!gsloop || !gsloop_con)
		return -EINVAL;

	err = rn_del(&gsloop->rn, gsloop_server->fd);
	if (err) {
		err = -err;	/* rn returns errno */
		dprintf(LEVEL_ERR, "rn_del failed, err = %d\n", err);
	}

	err = shutdown(gsloop_server->fd, SHUT_RDWR);
	if (err) {
		err = -errno;
		dprintf(LEVEL_ERR, "shutdown failed, err = %d\n", err);
	}

	err = close(gsloop_server->fd);
	if (err) {
		err = -errno;
		dprintf(LEVEL_ERR, "close failed, err = %d\n", err);
	}

	return err;
}


static int gsloop_inet_client_close(struct gsloop	*gsloop,
				    union gsloop_con	*gsloop_con)
{
	struct gsloop_client	*gsloop_client = &gsloop_con->client;
	int			err;

	err = rn_del(&gsloop->rn, gsloop_client->fd);
	if (err) {
		err = -err;	/* rn returns errno */
		dprintf(LEVEL_ERR, "rn_del failed, err = %d\n", err);
	}

	err = shutdown(gsloop_client->fd, SHUT_RDWR);
	if (err) {
		err = -errno;
		dprintf(LEVEL_ERR, "shutdown failed, err = %d\n", err);
	}

	err = close(gsloop_client->fd);
	if (err) {
		err = -errno;
		dprintf(LEVEL_ERR, "close failed, err = %d\n", err);
	}

	return err;
}


int gsloop_tcp_server_create(gsloop_t			*gsloop_anon,
			     uint16_t			port,
			     gsloop_client_cb_fn_t	client_cb,
			     void			*data,
			     gsloop_server_attr_t	*attr)
{
	struct gsloop		*gsloop = (struct gsloop *)gsloop_anon;
	struct gsloop_server	*server;
	int	err;
	int	fd;

	fd = gsloop_inet_server_open(SOCK_STREAM, port);
	if (fd < 0) {
		err = fd;
		goto exit_0;
	}

	server = &gsloop_con_alloc_and_add(gsloop, fd)->server;
	if (!server) {
		err = -ENOMEM;
		goto exit_1;
	}

	server->client_data	= data;
	server->close		= gsloop_inet_server_close;
	server->client_close	= gsloop_inet_client_close;
	server->client_cb	= client_cb;
	
	err = rn_prepare_fd_for_add(fd, getpid());
	if (err) {
		err = -err;	/* rn returns errno */
		goto exit_2;
	}

	err = rn_add(&gsloop->rn, fd, gsloop_server_callback, gsloop);
	if (err) {
		err = -err;
		goto exit_3;
	}
		
	return 0;

 exit_3:
 exit_2:
	/* FIXME: gsloop_con_free(server); */
 exit_1:
	close(fd);
 exit_0:
	return err;
}
